# SPDX-FileCopyrightText: 2022 RANCH Computing <contact@ranchcomputing.com>
# SPDX-License-Identifier: GPL-3.0-only

bl_info = {
    "name": "RANCHecker For Blender",
    "description": "Check your scene compatibility with the RANCH RenderFarm and prepare the archive to send on the RANCH",
    "author": "Ranch Computing",
"version": (1,2,10),
    "blender": (2, 93, 0),
    "location": "View3D",
    "wiki_url": "https://doc.ranchcomputing.com/blender:start",
    "tracker_url": "https://www.ranchcomputing.com/support/help-center",
    "category": "RANCH Computing",
}

# bl_info must be at the very top
# pylint: disable=wrong-import-position
import hashlib
import json
import logging
import os
import platform
import re
import shutil
import socket
import ssl
import sys
import tempfile
import time
import urllib.request
import zipfile

import bpy
import bpy.utils.previews

from bpy.props import (
    StringProperty,
    BoolProperty,
    IntProperty,
    PointerProperty,
    CollectionProperty,
    EnumProperty,
)

# pylint: enable=wrong-import-position

sys.path.append(os.path.dirname(__file__))
ssl._create_default_https_context = ssl._create_unverified_context

HOST = "127.0.0.1"
PORT = 7009

RANCHECKER_LOG = "RANCHecker.log"

"""
        https://devtalk.blender.org/t/enumproperty-and-string-encoding/7835
        10 june 2019 - EnumProperty and string encoding :
            There is a known bug with using a callback,
            Python must keep a reference to the strings returned by the callback
            or Blender will misbehave or even crash.
"""
priority_items = []  #


def _add_priorityFromColl_cb(self, context):
    """
    Priority Enum item callback
    Fill enum items with priority collection in scene
    """

    global priority_items
    priority_items = []

    scene = context.scene

    for item in scene.coll_priority_items:
        identifier = str(item.priority_id)
        name = item.priority_str
        description = str(item.priority_id)

        priority_items.append((identifier, name, description))

    return priority_items


def get_desktop():
    try:
        if os.path.isdir(os.path.join(os.path.expanduser("~"), "Desktop")):
            desktop = os.path.join(os.path.expanduser("~"), "Desktop")
        else:
            desktop = os.path.expanduser("~")
    except:
        pass
    return desktop


class SettingsProp(bpy.types.PropertyGroup):
    destination_bool: BoolProperty(name="Destination", description="Saving archive directory")
    str_destination: StringProperty(name="", default=get_desktop(), subtype="DIR_PATH")
    ftp_login: StringProperty(name="FTP Login", default="")
    ftp_password: StringProperty(name="FTP Password", default="")
    archive_path: StringProperty(name="archive path", default="")
    message: StringProperty(name="dialog message", default="")
    archive_done: BoolProperty(name="Archive done", default=False)

    priority: EnumProperty(
        name="priority",
        description="Priority to use on the RANCH",
        items=_add_priorityFromColl_cb,
    )

    @classmethod
    def register(cls):
        bpy.types.Scene.ranch_settings = PointerProperty(type=cls)

    @classmethod
    def unregister(cls):
        del bpy.types.Scene.ranch_settings


class ItemPriority(bpy.types.PropertyGroup):
    @classmethod
    def register(cls):
        bpy.types.Scene.coll_priority_items = bpy.props.CollectionProperty(type=ItemPriority)

    @classmethod
    def unregister(cls):
        del bpy.types.Scene.coll_priority_items

    priority_str: bpy.props.StringProperty(name="priority_str", default="")
    priority_id: bpy.props.IntProperty(name="priority_id", default=0)


# splitpath supports "/" and "\"
def splitpath(s):
    i = max(s.rfind("/"), s.rfind("\\"))
    return s[:i], s[i + 1 :]


def md5_basename(folder):
    """
    Compute a folder name (basename + md5 of the full path).
    Any update to this function must be copied to RANCHSet/RANCHecker.
    """
    # find part of path without "."
    parents, basename, i = folder, "", 0
    while parents and (not basename or "." in basename):
        parents, basename = splitpath(parents)
        if i > 1000:  # prevent infinite loop
            print("infinite splitpath on", folder)
            raise Exception("infinite splitpath on: " + folder)
        i += 1
    return basename + "_" + hashlib.md5(folder.encode()).hexdigest()


class Utils:
    forbidden_chars = re.compile("[^a-zA-Z0-9-_]+")
    RENDERMAN_OUTPUT_FORMATS = ("exr", "png", "tga", "tif")  # Default exr

    def __init__(self):
        pass

    @staticmethod
    def additional_folders(log):
        # Search for cache files, vdb, textures: images sequences or udim
        # Return (dirname, filepath), function
        for object in bpy.data.objects:
            # Not implemented yet: subscenes
            if object.library:
                log.warning(f" Subscene not yet supported: {object.library.filepath}")
                continue
            # cache
            for modifier in object.modifiers:
                try:
                    yield modifier.domain_settings.cache_directory, None
                except AttributeError:
                    pass

            # vdb must be manually handled
            # because bpy.ops.file.pack_all() does not work with vdb sequence
            # (only the first is packed)
            if object.type == "VOLUME":
                # take dirname (since filepath points to the first .vdb file)
                folder, _ = splitpath(object.data.filepath)
                yield folder, None

            # Textures : images sequence or UDIM
            if object.type == "MESH":
                for mat_slot in object.material_slots:
                    try:
                        for node in mat_slot.material.node_tree.nodes:
                            if node.type != "TEX_IMAGE" or not node.image.filepath:
                                continue

                            folder, filename = splitpath(node.image.filepath)
                            filename = re.escape(filename)

                            def replaceDigitSequence(repl, string):
                                return re.sub(
                                    r"\d{2,}", lambda _: repl, string
                                )  # lambda needed to prevent KeyError

                            if node.image.source == "TILED":  # UDIM
                                # https://docs.blender.org/manual/en/3.2/modeling/meshes/uv/workflows/udims.html?highlight=udim#file-substitution-tokens
                                pattern = filename.replace("<UDIM>", "\d+").replace("<UVTILE>", "u\d+_v\d+")
                                if pattern == filename:  # <...> not found, try other tricks
                                    pattern = replaceDigitSequence(
                                        "(\d+)|(u\d+_v\d+)", filename
                                    )  # detect digit sequence

                            elif node.image.source == "SEQUENCE":
                                pattern = replaceDigitSequence("\d+", filename)  # detect digit sequence
                            else:
                                print("unsupported TEX_IMAGE", node.image.source)
                                continue

                            pattern = re.compile(pattern)

                            def ignore_path(root, fileNames):
                                ignored = []
                                for f in fileNames:
                                    if pattern.match(f):
                                        continue
                                    ignored.append(f)
                                return ignored

                            yield (folder, ignore_path)
                    except AttributeError:
                        pass

        # cache (blender 2.80)
        try:
            yield bpy.context.object.filepath, None
        except AttributeError:
            pass

    @staticmethod
    def copy_additional_folders(folders, destination, log):
        """
        folders = list of (folder, ignore_function)
        """

        for folder, ignore_function in folders:
            if folder == "":
                continue

            if folder.startswith("//") and ".." not in folder:
                # relative path, without ".." : keep the same structure
                dst = folder[2:]
            else:
                # absolute path
                # find part of path without "."
                # append md5 to basename

                dst = md5_basename(folder)

            src = os.path.abspath(bpy.path.abspath(folder))
            dst = os.path.join(destination, dst)
            os.makedirs(dst, exist_ok=True)
            try:
                shutil.copytree(src=src, dst=dst, dirs_exist_ok=True, ignore=ignore_function)
            except FileNotFoundError:
                log.warning(f"Unable to copy '{src}': folder not found")

    @staticmethod
    def convert_format(file_format):
        static_formats = {
            "AVI_JPEG": "AVI_JPEG",
            "AVI_RAW": "AVI_RAW",
            "BMP": "bmp",
            "CINEON": "cin",
            "DPX": "dpx",
            "FFMPEG": "FFMPEG",
            "HDR": "hdr",
            "IRIS": "rgb",
            "JPEG": "jpg",
            "JPEG2000": "jpg",
            "OPEN_EXR_MULTILAYER": "exr",
            "OPEN_EXR": "exr",
            "PNG": "png",
            "TARGA_RAW": "tga",
            "TARGA": "tga",
            "TIFF": "tif",
            "WEBP": "webp",
        }

        return static_formats.get(file_format, "unknown")

    @staticmethod
    def make_dir(dir_name):
        if not os.path.isdir(dir_name):
            try:
                os.makedirs(dir_name, exist_ok=True)
            except Exception as e:
                logging.error(" {}".format(e))

    @staticmethod
    def remove_temp(temp_path):
        try:
            shutil.rmtree(temp_path)
        except Exception as e:
            logging.error(" {}".format(e))

    @staticmethod
    def creat_tempdir(desktop, proj_name):
        desktop_RANCH = os.path.join(desktop, "_RANCHecker_" + proj_name)
        status_dir = True
        while os.path.exists(desktop_RANCH):
            Utils.remove_temp(desktop_RANCH)
            time.sleep(0.1)
        Utils.make_dir(desktop_RANCH)
        return desktop_RANCH

    @staticmethod
    def save_as(desktop_temp, proj_name):
        new_proj = os.path.abspath(os.path.join(desktop_temp, proj_name))  # blender 3.1 needs absolute path
        save_back = bpy.context.preferences.filepaths.save_version
        try:
            bpy.context.preferences.filepaths.save_version = 0
            bpy.ops.wm.save_as_mainfile(filepath=new_proj, copy=True, compress=False, relative_remap=False)
            bpy.context.preferences.filepaths.save_version = save_back
        except Exception as e:
            print("could not save_as", e)
            pass

    @staticmethod
    def create_archive(desktop_temp, proj_name):
        name_only, ext_only = os.path.splitext(proj_name)
        archive_name = name_only + ".vub"
        zip_path = None
        try:
            dst = os.path.normpath(os.path.join(desktop_temp, "../"))
            with zipfile.ZipFile(
                os.path.join(dst, archive_name),
                "w",
                zipfile.ZIP_DEFLATED,
                allowZip64=True,
            ) as zip_path:
                for root, dirs, files in os.walk(desktop_temp):
                    for _file in files:
                        zip_filename = os.path.join(root, _file).replace(desktop_temp, "")
                        zip_path.write(os.path.join(root, _file), arcname=zip_filename)
            logging.info(" VUB Archive successfully created !")
            return archive_name
        except Exception as e:
            logging.error(" Archive Creation failed : " + str(e))
            return False

    @staticmethod  # ----- RENDERMAN ----
    def create_archive_using_renderman_packer(desktop_temp, proj_name):
        """
        Creates the .vub using the bpy.ops.renderman.scene_package operator
        """
        name_only, ext_only = os.path.splitext(proj_name)
        archive_name = name_only + ".vub"

        try:
            dst = os.path.normpath(os.path.join(desktop_temp, "../"))
            file_path = os.path.join(dst, archive_name)

            # Renderman packing in a dedicated directory (must be empty)
            rm_pack_dir = os.path.join(desktop_temp, "rm_pack")
            os.makedirs(rm_pack_dir, exist_ok=True)

            bpy.ops.renderman.scene_package(directory=rm_pack_dir, filepath=file_path)
            # exemple :
            #  bpy.ops.renderman.scene_package(
            #   directory="C:\\Users\\Ranch\\Desktop\\simpleTest\\RM_pack\\",
            #   filepath="C:\\Users\\Ranch\\Desktop\\simpleTest\\RM_pack\\TestBasicOpt3_udim.zip")

            logging.info(" VUB Archive successfully created !")

            return file_path
        except Exception as e:
            logging.error(" Archive Creation failed : " + str(e))
            return False

    @staticmethod
    def get_renderer(scene):
        renderer = scene.render.engine
        device = "CPU"
        if renderer == "CYCLES" and scene.cycles.device == "GPU":
            device = "GPU"
        elif renderer in ("BLENDER_EEVEE", "octane"):
            device = "GPU"
        return renderer, device

    @staticmethod
    def get_output_format(scene):
        _format = Utils.convert_format(scene.render.image_settings.file_format)

        if scene.render.engine == "PRMAN_RENDER":
            if _format not in Utils.RENDERMAN_OUTPUT_FORMATS:
                _format = "exr"

        elif _format == "unknown":
            _format = "png"
        return _format

    @staticmethod
    def get_output_path(scene):
        if scene.render.filepath.strip(" ") == "":
            return ""
        else:  #  [!] os.path.abspath("") renvoie le path de l'executable
            return os.path.abspath(bpy.path.abspath(scene.render.filepath.strip(" ")))

    @staticmethod
    def get_output_name(scene, log):
        output_path = Utils.get_output_path(scene)

        _format = Utils.get_output_format(scene)
        outName = Utils.get_scene_name(scene) + "_output"  # default name

        if (
            scene.render.filepath.strip(" ") == ""
            or scene.render.filepath.startswith("/tmp")
            or os.path.isdir(output_path)
        ):
            filepath = bpy.data.filepath  # .blend file name
            if filepath:  # .blend file name
                outName = os.path.splitext(os.path.basename(filepath))[0]
        else:
            basename = os.path.splitext(os.path.basename(output_path))[0]
            if basename:
                outName = basename

        if log and Utils.forbidden_chars.search(outName):
            log.warning(f"Special chars in outname not supported, will be replaced with '_': {outName}")

        file_name = Utils.forbidden_chars.sub("_", outName)
        return file_name + "_." + _format.lower()

    @staticmethod
    def get_resolution(scene):
        render_scale = Utils.get_render_scale(scene)
        res_x = int((scene.render.resolution_x * int(render_scale)) / 100)
        res_y = int((scene.render.resolution_y * int(render_scale)) / 100)
        return str(res_x), str(res_y)

    @staticmethod
    def get_strtend_frame(scene):
        start_frame = str(scene.frame_start)
        end_frame = str(scene.frame_end)
        return start_frame, end_frame

    @staticmethod
    def get_scene_name(scene):
        return scene.name_full

    @staticmethod
    def get_frame_step(scene):
        frame_step = scene.frame_step
        return frame_step

    @staticmethod
    def get_render_scale(scene):
        render_scale = scene.render.resolution_percentage
        return render_scale

    @staticmethod
    def get_destination(scene):
        scn_setting = scene.ranch_settings
        proj_name = os.path.basename(bpy.data.filepath)
        name_only, ext_only = os.path.splitext(proj_name)

        if scn_setting.destination_bool:
            # BUG if relative utiliser ->  os.path.abspath(bpy.path.abspath(folder))
            # des_temp = scn_setting.str_destination
            des_temp = os.path.abspath(bpy.path.abspath(scn_setting.str_destination))
            desktop_temp = Utils.creat_tempdir(des_temp, name_only)
            return desktop_temp
        else:
            desktop = get_desktop()
            desktop_temp = Utils.creat_tempdir(desktop, name_only)
            return desktop_temp

    @staticmethod
    def get_list_scenes(scene):
        scenes = None
        scn_set = scene.ranch_settings
        if scn_set.multiscene_bool:
            scenes = bpy.data.scenes
        else:
            scenes = scene
        return scenes

    @staticmethod
    def get_RANCHecker_version():
        return str(bl_info["version"][0]) + "." + str(bl_info["version"][1]) + "." + str(bl_info["version"][2])

    @staticmethod
    def get_multipass(scene, setpass):
        state = False
        try:
            nodes_name = []
            nodes = scene.node_tree.nodes
            for n in nodes:
                nodes_name.append(n.name)
            for node in nodes_name:
                if scene.node_tree.nodes[node].bl_idname == "CompositorNodeOutputFile":
                    if setpass:
                        bpy.context.scene.node_tree.nodes[node].base_path = "C:\\Blender\\Output\\"
                    state = True
        except:
            pass
        return state

    @staticmethod
    def set_log(string):
        if platform.system() == "Darwin":
            return string + "\r"
        elif platform.system() == "Windows":
            return string

    @staticmethod
    def unicode_filename(file_name):
        (root, _) = os.path.splitext(file_name)
        return not Utils.forbidden_chars.search(root)


class RANCH_OT_message(bpy.types.Operator):
    """
    Draw one message line in UI dialog
    """

    bl_idname = "ranch.message"
    bl_label = "Ranch message:"

    def execute(self, context):
        try:
            bpy.ops.wm.save_mainfile(filepath=bpy.data.filepath)
            return {"FINISHED"}
        except:
            return {"FINISHED"}

    def invoke(self, context, event):
        wm = context.window_manager
        return wm.invoke_props_dialog(self, width=450)

    def draw(self, context):
        ranch_setting = bpy.context.scene.ranch_settings
        layout = self.layout
        row = layout.row()
        row.label(text=ranch_setting.message, icon="INFO")


# taken from RANCHSync repo
def send_to_ranchsync(data: object) -> object:
    def recvall(sock):
        BUFF_SIZE = 4096  # 4 KiB
        data = b""
        while True:
            part = sock.recv(BUFF_SIZE)
            data += part
            if len(part) < BUFF_SIZE:
                # either 0 or end of data
                break
        return data

    HOST = "127.0.0.1"
    PORT = 7009
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((HOST, PORT))
        json_data = json.dumps(data)
        s.sendall(bytes(json_data, encoding="utf-8"))
        b = recvall(s)
        return json.loads(str(b, "utf-8"))
    finally:
        s.close()


class UnsupportedBlenderVersion(Exception):
    def __init__(self, version):
        super().__init__(f"Blender blender v{version} is not supported (yet), please contact us.")
        self.version = version


class SubmitJob:
    def get_supported_renderers(self):
        blender_version = str(bpy.app.version[0]) + "." + str(bpy.app.version[1])
        for release in RANCH_OT_sweb_status.supported_releases():
            if blender_version == str(release["version"]):
                return release["renderers"]
        raise UnsupportedBlenderVersion(blender_version)

    def get_sweb_slug(self, scene):
        scene_renderer = scene.render.engine.lower()
        if scene_renderer == "blender_eevee":
            scene_renderer = "eevee"
        elif scene_renderer == "cycles":
            scene_renderer += "-" + scene.cycles.device.lower()
        return scene_renderer

    def get_sweb_rendererID(self, scene_renderer):
        renderers = self.get_supported_renderers()
        for renderer in renderers:
            if scene_renderer == renderer["slug"]:
                return renderer["id"]
        return 0

    def __get_sweb_priority(self):
        """
        send a request to sweb to retreive priorities for current renderer
        return list or false
        """
        scene_renderer = self.get_sweb_slug(bpy.context.scene)
        if not scene_renderer:
            return False
        renderer_id = self.get_sweb_rendererID(scene_renderer)
        print("render_id = ", renderer_id)

        legacy_farm_filter = ""
        if not renderer_id and scene_renderer.startswith("cycles-"):
            # try again (sweb has not been migrated yet)
            renderer_id = self.get_sweb_rendererID("cycles")
            legacy_farm_filter = scene_renderer[len("cycles-") :]
            if legacy_farm_filter == "cpu":
                legacy_farm_filter = "power"

        if not renderer_id:
            bpy.context.scene.ranch_settings.message = (
                "Renderer {} not supported (yet), please contact us.".format(scene_renderer)
            )
            print("Renderer not supported: " + scene_renderer)
            return False
        data = {
            "method": "sweb_get_json",
            "url": "api/renderer-releases/{}/priorities".format(renderer_id),
        }
        resp = self.send_to_ranchsync(data)
        if not resp:
            return False

        for farm, priorities in resp["farms"].items():
            if legacy_farm_filter and farm != legacy_farm_filter:
                continue
            yield from reversed(priorities)

    def fill_coll_priorities(self):
        """
        Retreive priorities from RANCH site
        if successfully retreived,  fill priorities scene collection
        return len(collection)
        """

        """
            First clear collection
        """
        if len(bpy.context.scene.coll_priority_items) != 0:
            bpy.context.scene.coll_priority_items.clear()
        bpy.context.scene.ranch_settings.message = ""

        priorities = self.__get_sweb_priority()  # Send a request to sweb to get RANCH priorities

        if priorities:
            for p in priorities:
                item = ItemPriority(bpy.context.scene.coll_priority_items.add())
                item.priority_id = int(p["id"])
                item.priority_str = (
                    p["name"]
                    + " /Price "
                    + str(p["ghz_h_price"])
                    + " /"
                    + str(p["max_allocated_nodes"])
                    + " "
                    + str(p["nodes_unit"])
                )

        else:
            if bpy.context.scene.ranch_settings.message == "":
                bpy.context.scene.ranch_settings.message = "Please run RANCHSync if you want to submit."
            print(" WARNING : Can't prepare submission : internet not connected or RANCHSync not running.")
            bpy.ops.ranch.message("INVOKE_DEFAULT")
            return 1

        return len(bpy.context.scene.coll_priority_items)

    def submit_sync(self):
        ranch_setting = bpy.context.scene.ranch_settings
        archive_path = ranch_setting.archive_path

        if not archive_path or not os.path.exists(archive_path):
            ranch_setting.message = "RANCHSync unable to submit: Archive not created or errors during the process"
            return 1
        if len(bpy.context.scene.coll_priority_items) == 0:
            ranch_setting.message = "Unable to get priorities: RANCHSync not running or Renderer not suported yet"
            return 1

        scene_renderer = self.get_sweb_slug(bpy.context.scene)
        renderer_id = self.get_sweb_rendererID(scene_renderer)
        if not renderer_id and scene_renderer.startswith("cycles-"):
            # try again with "cycles" (sweb has not been migrated yet)
            renderer_id = self.get_sweb_rendererID("cycles")

        if not renderer_id:
            ranch_setting.message = "Renderer not supported: " + scene_renderer
            return 1

        download_param = {"remote_subfolder": "Images"}
        local_path = Utils.get_output_path(bpy.context.scene)
        if local_path:
            if os.path.isdir(local_path):
                download_param["local_folder"] = local_path
            else:
                download_param["local_folder"] = os.path.dirname(local_path)

        resp = self.send_to_ranchsync(
            {
                "method": "create_and_upload",
                "sweb": {
                    "renderer_id": str(renderer_id),
                    "priority_id": str(ranch_setting.priority),
                    "disclaimer": True,
                },
                "upload": {
                    "file_path": archive_path,
                },
                "download": download_param,
            }
        )
        if not resp:
            return 0
        # print("resp data ", resp)
        return 2

    def send_to_ranchsync(self, data):
        try:
            resp = send_to_ranchsync(data)
            if "error" in resp:
                bpy.context.scene.ranch_settings.message = resp["error"]
                return False
            return resp
        except ConnectionRefusedError as a:
            bpy.context.scene.ranch_settings.message = "Please run RANCHsync V4 to start uploading."
            return False


class RANCH_OT_submit(bpy.types.Operator):
    """
    Submit Job to RANCH (via RANCHSync/sweb)
    """

    bl_idname = "ranch.submit"
    bl_label = "Submit"
    bl_description = "Submit your project on the RANCH site"

    def execute(self, context):
        ranch_setting = bpy.context.scene.ranch_settings
        ranch_setting.message = "Unknown error"
        create = SubmitJob()
        job = create.submit_sync()
        if job == 1:
            bpy.ops.ranch.message("INVOKE_DEFAULT")
            return {"FINISHED"}
        elif job == 2:
            self.report({"INFO"}, "Job uploaded ...")

            return {"FINISHED"}
        else:
            bpy.ops.ranch.message("INVOKE_DEFAULT")

        return {"FINISHED"}


class SetupLogging:
    def __init__(self):
        self.file_handler = None
        self.log = None

    def setup(self, desktop_temp, filename, Formatter):
        logging_ini = os.path.join(desktop_temp, filename)
        formatter = logging.Formatter(Formatter)
        self.log = logging.getLogger(__name__)
        self.log.setLevel(logging.DEBUG)
        self.file_handler = logging.FileHandler(logging_ini, encoding="UTF-8")
        self.file_handler.setFormatter(formatter)
        self.log.addHandler(self.file_handler)

    def close_log(self):
        self.file_handler.close()
        self.log.removeHandler(self.file_handler)


class LogChecks(SetupLogging):
    def __init__(self, desktop_temp):
        self.scene = bpy.context.scene
        self.desktop_temp = desktop_temp
        self.setup(self.desktop_temp, RANCHECKER_LOG, "%(levelname)s:%(message)s")
        self.__ranchecker_version()
        self.__blender_version()
        self.all_checks = [
            self.__resolution_rercentage,
            self.__frame_step,
            self.__render_scale,
            self.__render_region,
            self.__stereoscopy,
            self.__multipass,
            self.__video_format,
            self.__unknown_format,
            self.__pack_scene,  # last
        ]

    def __ranchecker_version(self):
        version = Utils.get_RANCHecker_version()
        self.log.info("RANCHecker version : {}".format(version))

    def __blender_version(self):
        self.log.info("Blender version : {}".format(bpy.app.version))

    def __resolution_rercentage(self):
        res = Utils.get_render_scale(self.scene)
        self.log.info("Scale percentage : {}%".format(res))

    def __pack_scene(self):
        if not bpy.data.use_autopack:
            self.log.info("FYI - RANCH Rendering: use File->External Data->Automatically Pack Ressources")

    def __frame_step(self):
        frame_step = Utils.get_frame_step(self.scene)
        if frame_step > 1:
            # Error : ("<!> Frame step > 1 NOT ALLOWED, please go to our submission web page.")
            self.log.error(
                "<!> Frame step > 1 NOT ALLOWED, please go to our submission web page to submit only specific frames."
            )

    def __render_scale(self):
        render_scale = Utils.get_render_scale(self.scene)
        if render_scale != 100:
            self.log.warning(
                "Percentage scale is {}%, your render resolution is not at full size".format(render_scale)
            )

    def __render_region(self):
        region = self.scene.render.use_border
        if region:
            self.log.warning("Render region is active")

    def __stereoscopy(self):
        stereo = self.scene.render.use_multiview
        if stereo:
            self.log.warning("Stereoscopy is enabled")

    def __multipass(self):
        if Utils.get_multipass(self.scene, False):
            self.log.warning("Multipasses output : {} ".format("Enable"))

    def __video_format(self):
        videoFormats = ["AVI_JPEG", "AVI_RAW", "FFMPEG"]
        file_format = Utils.get_output_format(self.scene)
        if file_format in videoFormats:
            self.log.error(
                "[ {} ] videos are not supported on the farm, only images sequence. Please use image format".format(
                    file_format
                )
            )

    def __unknown_format(self):
        scene_outformat = self.scene.render.image_settings.file_format
        _format = Utils.convert_format(scene_outformat)

        if self.scene.render.engine == "PRMAN_RENDER":
            if scene_outformat == "OPEN_EXR_MULTILAYER" or _format not in Utils.RENDERMAN_OUTPUT_FORMATS:
                self.log.warning(
                    "[ {} ] not yet supported, forced to exr single layer. \
                    Please contact us contact@ranchcomputing.com".format(
                        scene_outformat
                    )
                )
        elif _format == "unknown":
            self.log.warning(
                "[ {} ] not yet supported, forced to png. Please contact us contact@ranchcomputing.com".format(
                    scene_outformat
                )
            )

    def reset_scene(self):
        self.scene = bpy.context.scene

    def write_log_file(self):
        for check in self.all_checks:
            try:
                check()
            except:
                pass
        self.close_log()


class CompanionFileContent:
    def __init__(self, scene, log):
        self.scene = scene
        self.log = log

    def outputs_lists(self):
        output_name = Utils.get_output_name(self.scene, self.log)
        yield output_name

    def ranchecker_info(self):
        renderer, device = Utils.get_renderer(self.scene)
        render_engine = renderer + "-" + device
        yield render_engine
        self.log.info("Render engine : {}".format(render_engine))
        if renderer == "BLENDER_EEVEE":
            self.log.warning(
                "EEVEE is not multi GPU, your price estimation may be inaccurate. Please submit a few frames first."
            )

        output_name = Utils.get_output_name(self.scene, None)
        # BUG Blender RenderMan 24.4:
        #   - blender.exe command line -o -> produce black frame
        #     -> redirected to  "C:\\Blender\\Output\\renderman-black-glitch\\" (dans .info)
        #   - RenderMan output -> redirected  to "C:\\Blender\\Output\\" in RanchSet
        if renderer == "PRMAN_RENDER":
            output_path = "C:\\Blender\\Output\\renderman-black-glitch\\" + output_name
        else:
            output_path = "C:\\Blender\\Output\\" + output_name
        yield output_path
        self.log.info("Output format : {}".format(output_name))

        _x, _y = Utils.get_resolution(self.scene)
        resolution = _x + " " + "x" + " " + _y
        yield resolution
        self.log.info("Resolution : {} X {}".format(_x, _y))

        start_frame, end_frame = Utils.get_strtend_frame(self.scene)
        frame_range = start_frame + " " + "-" + " " + end_frame
        yield frame_range
        self.log.info("Frame range : {} - {}".format(start_frame, end_frame))

        scene_name = Utils.get_scene_name(self.scene)
        yield scene_name
        self.log.info("Current scene : {}".format(scene_name))

        if self.scene.display_settings.display_device == "ACES":
            yield "ACES"
            self.log.info("Color Management : ACES")
        else:
            yield ""
            self.log.info(
                "Color Management : {} - {}".format(
                    self.scene.display_settings.display_device, self.scene.view_settings.view_transform
                )
            )


class RANCH_OT_create_archive(bpy.types.Operator):
    """
    Create archive (.vub) for the RANCH
    """

    bl_idname = "ranch.create_archive"
    bl_label = "Create Archive"
    bl_description = "Create your .vub archive for the RANCH"

    def __init__(self):
        self.file_path = bpy.data.filepath
        filename, ext = os.path.splitext(os.path.basename(bpy.data.filepath))
        self.proj_name = Utils.forbidden_chars.sub("_", filename) + ext
        self.name_only, ext_only = os.path.splitext(self.proj_name)
        if bpy.data.filepath == "":  # empty bl filepath means the file is not saved
            self.file_path = "untitled"
            self.proj_name = "untitled.blend"
            self.name_only = "untitled"

        self.desktop_temp = None
        self.create_state = True

    def execute(self, context):
        scn = context.scene

        self.desktop_temp = Utils.get_destination(scn)
        print("self.desktop_temp ", self.desktop_temp)
        self.log_data = LogChecks(self.desktop_temp)

        # attempt to pack all assets
        self.pack_all(self.log_data.log)

        # save as, scene may be reloaded : actualize handler
        scn = self.savescene_in_tmpdir(scn)

        vub_filepath = ""
        renderer, device = Utils.get_renderer(scn)

        #       --- RENDERMAN ---
        if bpy.context.scene.render.engine == "PRMAN_RENDER":
            try:
                vub_filepath = Utils.create_archive_using_renderman_packer(self.desktop_temp, self.proj_name)
            except Exception as e:
                self.log_data.log.error(str(e))
            # Write  .info, .txt in vub
            self.write_companion_files_to_zip(vub_filepath, self.log_data.log)
            print("  ------RENDERMAN PACKING ----- ")

        else:
            folders = Utils.additional_folders(self.log_data.log)
            self.write_companion_files_to_folder(self.desktop_temp, self.log_data.log)
            Utils.copy_additional_folders(folders, self.desktop_temp, self.log_data.log)
            vub_filepath = Utils.create_archive(self.desktop_temp, self.proj_name)

        # scene may be reloaded : actualize handler
        scn = context.scene
        self.log_data.reset_scene()

        if not vub_filepath:  # Early return with error
            self.create_state = False
        if not self.create_state:  # self.create_state will be False if an error was logged or no archive created
            self.report({"ERROR"}, "Failed to create archive")
            return {"FINISHED"}

        self.archive_success(scn)

        # run all_checks
        self.log_data.write_log_file()
        # display log messages in UI
        scn.ranch_settings.archive_path = self.display_log()

        # Write log at the end in vub
        self.write_log_to_zip(vub_filepath)

        # Last : remove temporary directory
        Utils.remove_temp(self.desktop_temp)

        return {"FINISHED"}

    def archive_success(self, scn):
        scn.ranch_settings.archive_done = True
        """ Archive created successfully, prepare submit : sweb request for priorities """
        sj = SubmitJob()
        try:
            sj.fill_coll_priorities()
        except Exception as e:
            self.log_data.log.warning("Unable to get priorities on RANCH: " + str(e))

        if scn.ranch_settings.message:
            self.log_data.log.warning(scn.ranch_settings.message)
            print("ranch_settings.message = ", scn.ranch_settings.message)
            self.report({"WARNING"}, scn.ranch_settings.message)

            # - to solve -
            # Handlers are cleared when bpy.ops.wm.revert_mainfile()/save/saveas is called
            # (INVOKE_DEFAULT doesn't work)-> EXEC_DEFAULT
            # https://blender.stackexchange.com/questions/6101/
            # poll-failed-context-incorrect-example-bpy-ops-view3d-background-image-add/6105#6105
            bpy.ops.ranch.message("EXEC_DEFAULT")

    @staticmethod
    def pack_all(log):
        print("bpy.ops.file.pack_all")
        try:
            bpy.ops.file.pack_all()
        except RuntimeError as e:
            for msg in str(e).splitlines():
                log.warning(msg)

        print("bpy.ops.file.pack_libraries")
        try:
            bpy.ops.file.pack_libraries()
        except RuntimeError as e:
            for msg in str(e).splitlines():
                log.warning(msg)

    def savescene_in_tmpdir(self, scn):
        if bpy.data.is_dirty:
            # Changes have been made since the last save. Display message.
            scn.ranch_settings.message = " WARNING : The scene will be saved."
            self.log_data.log.warning("___ The scene has been saved.___")
            print(" WARNING :___ The scene has been saved.___")
            bpy.ops.ranch.message("INVOKE_DEFAULT")

        Utils.save_as(self.desktop_temp, self.proj_name)

        # save as : scene may be reloaded :
        # actualize log handler, return context.scene
        self.log_data.reset_scene()
        return bpy.context.scene

    def companion_files_contents(self, log):
        def join_lines(lines):
            return "\r\n".join(lines) + "\r\n"

        companionFileContent = CompanionFileContent(bpy.context.scene, log)
        yield ("Outputs_list.txt", join_lines(companionFileContent.outputs_lists()))
        yield ("RANCHecker.info", join_lines(companionFileContent.ranchecker_info()))

    def write_companion_files_to_folder(self, folder, log):
        for filename, content in self.companion_files_contents(log):
            with open(os.path.join(folder, filename), "w", newline="") as f:
                f.write(content)

    def write_companion_files_to_zip(self, archive_filepath, log):
        with zipfile.ZipFile(archive_filepath, "a", zipfile.ZIP_DEFLATED, allowZip64=True) as zip_path:
            for filename, content in self.companion_files_contents(log):
                zip_path.writestr(filename, content)

    def write_log_to_zip(self, archive_filepath):
        with zipfile.ZipFile(archive_filepath, "a", zipfile.ZIP_DEFLATED, allowZip64=True) as zip_path:
            zip_path.write(os.path.join(self.desktop_temp, "RANCHecker.log"), arcname="RANCHecker.log")

    def display_log(self):
        self.clear_list()
        log_file = os.path.join(self.desktop_temp, "RANCHecker.log")
        with open(log_file, "r") as _file:
            for line in _file:
                [state, message] = line.split(":", 1)
                item = bpy.context.scene.log_list.add()
                item.msg = message.replace("\n", "")
                if state == "INFO":
                    item.icon = "INFO"
                elif state == "ERROR":
                    item.icon = "ERROR"
                    self.create_state = False
                elif state == "WARNING":
                    item.icon = "CANCEL"
        if self.create_state:
            item = bpy.context.scene.log_list.add()
            dir_path = os.path.dirname(self.desktop_temp)
            archive_path = os.path.join(dir_path, self.name_only + ".vub")
            # self.ranch_setting.archive_path = archive_path
            item.msg = "Archive created : {}".format(archive_path)
            item.icon = "OUTLINER_OB_GROUP_INSTANCE"
            item = bpy.context.scene.log_list.add()
            item.msg = "=== Archive successfully created ==="
            item.icon = "KEYTYPE_JITTER_VEC"
            return archive_path
        else:
            item = bpy.context.scene.log_list.add()
            item.msg = "=== Failed to create archive (See Error detail in System Console) ==="
            item.icon = "ERROR"
            return ""

    def clear_list(self):
        logs = bpy.context.scene.log_list
        while len(logs):
            logs.remove(0)


class MultiSceneProp(bpy.types.PropertyGroup):
    scn_name: StringProperty(name="Name", description="item", default="Untitled")
    scn_state: bpy.props.BoolProperty(name="", default=False)


class RANCH_UL_logs_list(bpy.types.UIList):
    def draw_item(
        self,
        context,
        layout,
        data,
        item,
        icon,
        active_data,
        active_propname,
        index,
        flt_flag,
    ):
        if self.layout_type in {"DEFAULT", "COMPACT"}:
            if item.icon == "INFO":
                my_icon = "QUESTION"
            elif item.icon == "CANCEL":
                my_icon = "ERROR"
            elif item.icon == "ERROR":
                my_icon = "CANCEL"
            else:
                my_icon = "CHECKMARK"
            layout.prop(
                item,
                "msg",
                text="",
                emboss=False,
                expand=False,
                translate=False,
                icon=my_icon,
            )

    def invoke(self, context, event):
        pass


class RANCH_OT_refresh_scenes(bpy.types.Operator):
    bl_idname = "ranch.refresh_scenes"
    bl_label = "Refresh scenes"
    add = bpy.props.BoolProperty(default=True)

    def execute(self, context):
        scns = []
        scns = bpy.data.scenes
        self.clear(context)
        for snc in scns:
            item = context.scene.ranch_scenes.add()
            item.scn_name = str(snc.name)
        return {"FINISHED"}

    def clear(self, context):
        scn = context.scene.ranch_scenes
        while len(scn):
            scn.remove(0)


class View3DPanel:
    bl_category = "RANCHECKER"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"

    @classmethod
    def poll(cls, context):
        # return context.object is not None -> Bug ? (context.object contient les objets actifs )
        return context != None


class RANCH_PT_main(View3DPanel, bpy.types.Panel):
    bl_label = "RANCHecker " + " v" + Utils.get_RANCHecker_version()

    def draw(self, context):
        scn_set = context.scene.ranch_settings
        layout = self.layout


class RANCH_PT_create_archive(View3DPanel, bpy.types.Panel):
    bl_parent_id = "RANCH_PT_main"
    bl_label = "Create Archive"

    def draw(self, context):
        scn_set = context.scene.ranch_settings
        layout = self.layout

        row = layout.row()
        row.prop(scn_set, "destination_bool")
        if scn_set.destination_bool:
            row = layout.row()
            row.prop(scn_set, "str_destination")

        row = layout.row()
        row.operator("ranch.create_archive", text="Create Archive", icon="PACKAGE")


class RANCH_PT_submit_job(View3DPanel, bpy.types.Panel):
    bl_parent_id = "RANCH_PT_main"
    bl_label = "Submit Job"

    def draw(self, context):
        scn_set = context.scene.ranch_settings
        layout = self.layout

        row = layout.row()

        row.prop(scn_set, "priority")
        row = layout.row()
        row.operator("ranch.submit", text="RANCH Submit", icon="EXPORT")


class LogProp(bpy.types.PropertyGroup):
    icon: StringProperty(name="", default="")
    msg: StringProperty(name="", default="")


class RANCH_OT_clear_logs(bpy.types.Operator):
    bl_idname = "ranch.clear_logs"
    bl_label = "Clear logs"
    bl_description = "Clear the log window"

    def execute(self, context):
        logs = context.scene.log_list
        while len(logs):
            logs.remove(0)
        return {"FINISHED"}


class RANCH_PT_logs(View3DPanel, bpy.types.Panel):
    bl_label = "Logs"

    def draw(self, context):
        layout = self.layout
        objs = context.scene
        layout.template_list("RANCH_UL_logs_list", "", objs, "log_list", objs, "log_list_index")
        layout.operator("ranch.clear_logs", icon="WORDWRAP_OFF")


class RANCH_PT_check_update(View3DPanel, bpy.types.Panel):
    bl_label = "RANCHecker update"

    def draw(self, context):
        layout = self.layout

        layout.row().operator(RANCH_OT_sweb_status.bl_idname, icon="FILE_REFRESH")

        for i, text in enumerate(RANCH_OT_sweb_status.status_lines):
            icon = "NONE"
            if i == 0 and RANCH_OT_sweb_status.icon:
                icon = RANCH_OT_sweb_status.icon
            row = layout.row()
            if i < len(RANCH_OT_sweb_status.status_lines) - 1:
                row.ui_units_y = 0.75  # smaller line height
            row.label(text=text, icon=icon)

        if RANCH_OT_sweb_status.ranchecker_url:
            layout.row().operator(RANCH_OT_upgrade_self.bl_idname, icon="IMPORT")


class RANCH_OT_sweb_status(bpy.types.Operator):
    """
    Retrieve the latest RANCHecker information online
    """

    bl_idname = "ranch.sweb_status"
    bl_label = "Check for update"
    bl_description = "Retrieve the latest RANCHecker information online"

    status_lines = []
    icon = ""
    _data = {}
    _fetched_timestamp = 0.0
    ranchecker_url = ""  # will be filled with the URL when an upgrade is available

    def execute(self, context):
        self.retrieve_online_info()
        return {"FINISHED"}

    @classmethod
    def retrieve_online_info(cls):
        cls.status_lines = ["Checking for an update..."]
        cls.icon = "FILE_REFRESH"
        cls.ranchecker_url = ""
        print("Retrieving latest RANCHecker information online...")
        try:
            # blender becomes unresponsive during this operator.
            # See https://blender.stackexchange.com/a/71830/118994 for possible solutions
            with urllib.request.urlopen(
                "https://www.ranchcomputing.com/api/ranchtools/ranchecker/blender.json"
            ) as response:
                cls._data = json.loads(response.read())
                cls._fetched_timestamp = time.time()
                cls.update_status_text(cls._data["ranchecker"])
        except Exception as e:
            cls.status_lines = [str(e)]
            cls.icon = "ERROR"

    @classmethod
    def update_status_text(cls, ranchecker):
        current_version = bl_info["version"]
        online_version = tuple(map(int, ranchecker["last_version"].split(".")))
        if current_version >= online_version:
            cls.status_lines = ["You are using the latest version."]
            cls.icon = ""
            return

        changelog_lines = ranchecker["last_changelog"].split("\r\n") if ranchecker["last_changelog"] else []
        cls.status_lines = ["Version " + ranchecker["last_version"] + " is available!"] + changelog_lines
        cls.icon = "INFO"
        cls.ranchecker_url = ranchecker["url"]

    @classmethod
    def _retrieve_online_info_if_stale(cls):
        if cls._fetched_timestamp > 0 and time.time() < cls._fetched_timestamp + 60 * 60:
            # online info was fetch successfully less than 1 hour ago: use the cached version
            return
        cls.retrieve_online_info()

    @classmethod
    def release_filename(cls) -> str:
        return cls._data["ranchecker"]["filename"]

    @classmethod
    def supported_releases(cls):
        cls._retrieve_online_info_if_stale()
        return cls._data["application"]["releases"]


class RANCH_OT_upgrade_self(bpy.types.Operator):
    """
    Download and install the latest RANCHecker version
    """

    bl_idname = "ranch.upgrade_self"
    bl_label = "Upgrade RANCHecker"
    bl_description = "Download and install the latest RANCHecker version"

    def execute(self, context):
        with tempfile.TemporaryDirectory(prefix="ranch-") as tmpdir:
            download_path = os.path.join(tmpdir, RANCH_OT_sweb_status.release_filename())
            urllib.request.urlretrieve(RANCH_OT_sweb_status.ranchecker_url, download_path)
            res = bpy.ops.preferences.addon_install(filepath=download_path)
            if res != {"FINISHED"}:
                RANCH_OT_sweb_status.icon = "ERROR"
                RANCH_OT_sweb_status.status_lines = ["Upgrade could not be installed!"]
                return res
            RANCH_OT_sweb_status.icon = "INFO"
            RANCH_OT_sweb_status.status_lines = ["Please restart blender", "to complete the upgrade."]
            RANCH_OT_sweb_status.ranchecker_url = ""
            return res


classes = (
    RANCH_PT_main,
    RANCH_PT_create_archive,
    RANCH_PT_submit_job,
    RANCH_OT_create_archive,
    RANCH_OT_submit,
    ItemPriority,
    RANCH_OT_message,
    SettingsProp,
    MultiSceneProp,
    RANCH_OT_refresh_scenes,
    LogProp,
    RANCH_OT_clear_logs,
    RANCH_PT_logs,
    RANCH_UL_logs_list,
    RANCH_PT_check_update,
    RANCH_OT_sweb_status,
    RANCH_OT_upgrade_self,
)

preview_collections = {}


def register():
    from bpy.utils import register_class

    for cls in classes:
        register_class(cls)

    bpy.types.Scene.ranch_scenes = CollectionProperty(type=MultiSceneProp)
    bpy.types.Scene.ranch_scenes_index = IntProperty(default=0)
    bpy.types.Scene.log_list = CollectionProperty(type=LogProp)
    bpy.types.Scene.log_list_index = IntProperty(default=0)


def unregister():
    from bpy.utils import unregister_class

    for pcoll in preview_collections.values():
        bpy.utils.previews.remove(pcoll)
    preview_collections.clear()
    for cls in classes:
        unregister_class(cls)
    del bpy.types.Scene.ranch_scenes
    del bpy.types.Scene.ranch_scenes_index


if __name__ == "__main__":
    register()

    if "doctest" in sys.argv:
        import doctest

        (failure_count, test_count) = doctest.testmod()
        print(test_count, "tests passed.")
        assert not failure_count, f"{failure_count} doctest failures"
